import List
import Ratio
import Debug.Trace

newtype Clause a = Clause ([a],[a]) deriving (Eq,Show)

class (Eq f,Show f) => Logic f where 
	ff :: f; tt :: f;
	x :: f; y :: f; z :: f;
	neg :: f -> f
	(/\) :: f -> f -> f
	(\/) :: f -> f -> f
	onlybox :: f -> [f]
	disj :: [f] -> f
	ma :: f -> [f]
	eval :: f -> Clause f -> Bool
	provable :: f -> Bool

class (Logic f,Logic g) => Strip f g where
	strip :: f -> [g]

data L1 = F1 | T1 | Atom1 Int | Neg1 L1 | And1 L1 L1 | Or1 L1 L1 | HM1 Char L1 deriving (Eq,Show)

instance Strip L1 L1 where
	strip (HM1 _ x) = [x]; strip _ = [];

instance Logic L1 where
	ff = F1; tt = T1;
	x = Atom1 1; y = Atom1 2; z = Atom1 3;
	neg F1 = T1; neg T1 = F1; neg x = Neg1 x
	x \/ y = Or1 x y
	x /\ y = And1 x y
	
	onlybox (HM1 ind x) = [HM1 ind x]; onlybox _ = []

	disj [] = F1; disj [a] = a; disj (a:as) = Or1 a (disj as)
	ma (And1 phi psi) = (ma phi) `union` (ma psi)
	ma (Or1 phi psi) = (ma phi) `union` (ma psi)
	ma (Neg1 phi) = ma phi
	ma x = [x]
	eval (And1 phi psi) s = (eval phi s) && (eval psi s)
	eval (Or1 phi psi) s = (eval phi s) || (eval psi s)
	eval (Neg1 phi) s  = not (eval phi s)
	eval T1 s = True; eval F1 s = False
	eval phi (Clause (pos, neg))
	  | elem phi neg = False
	  | elem phi pos = True
	
	provable phi = all (\c -> any (all provable) (match1 c)) (cnf phi) 


match1 :: Clause L1 -> [[L1]]
match1 (Clause (pl,nl)) =
  let isag a (HM1 ind x) = (ind == a); -- is it agent a?
      ag (HM1 ind x) = ind; -- find the agent
      ags = nub (map ag (pl ++ nl)); -- the set of agents we have
      aglits = map (\x -> (filter (isag x) pl, filter (isag x) nl)) ags;
      -- this is a list of pairs (pl, nl) for each agent
      strlits  = map (\(x, y) -> (map neg (striplst x), striplst y)) aglits;
      -- the same with boxes stripped
      comb (nls, pls) = map disjlst (map (\x -> x:nls) pls);
      -- turn a combination pls/nls into a list of formulas
      fllist = map comb strlits;
      -- a list of lists consisting of formulas
  in concat fllist


-- generic code (no need for duplication)
------------------------------------------

(-->) :: (Logic f) => f -> f -> f
x --> y = (neg x) \/ y

(<-->) :: (Logic f) => f -> f -> f
x <--> y = (x --> y) /\ (y --> x)

-- disjunction of negation of elements
ndisj :: (Logic f) => [f] -> f
ndisj xs = disj (map neg xs)


-- create singleton lists containing disjunctions
disjlst :: (Logic f) => [f] -> [f]
disjlst ls = [disj ls]

-- negate a list
neglst :: (Logic f) => [f] -> [f]
neglst xs = map neg xs 

-- generic: purely generic

-- avoids []->[[]], otherwise F matches F
listify :: [a] -> [[a]]
listify [] = []; listify x = [x]

-- length ordering on lists
cmpsize :: [a] -> [a] -> Ordering
cmpsize s t = if(length(s) < length(t)) then LT else GT

-- zip two lists together using a binary operator
zipbin :: (a -> a -> a) -> [a] -> [a] -> [a]
zipbin _ [] _ = [];	zipbin _ _ [] = [];
zipbin f (x:xs) (y:ys) = (f x y):(zipbin f xs ys)

-- given indices and a list find image
img :: [Int] -> [a] -> [a]
img inds ls = map (ls!!) inds 

-- map a list to indices from 0 -> length -1
toinds :: [a] -> [Int]
toinds [] = []; toinds xs = [0..(length xs)-1]


-- construct all integer n-tuples with elements from 1,..,r
tuprange :: Int -> Int -> [[Int]]
tuprange _ 0 = [[]]
tuprange r n = 
	let rec xs ys = map (\z -> z:ys) xs
	in	concatMap (rec [1..r]) (tuprange r (n-1))

	
-- generic: set operations

pwdisj :: (Eq a) => [[a]] -> Bool
pwdisj [] = True
pwdisj (x:xs) = if (all (\y -> null(x `intersect` y)) xs) then (pwdisj xs) else False
	
pwrset :: [a] -> [[a]]
pwrset [] = [[]]; pwrset (x:xs) = (map ([x]++) (pwrset xs)) ++ (pwrset xs) 

-- remove subsets
rmsub :: (Eq a) => [[a]] -> [[a]]
rmsub [] = []
rmsub (x:xs) | (any (\y -> (x `intersect` y == x)) xs) = rmsub(xs) | otherwise = x:rmsub(xs) 


-- generic: strippers and box collecting

-- strip a list
striplst :: (Strip f g) => [f] -> [g]
striplst ls = concatMap strip ls

onlyboxes :: (Logic f) => [f] -> [f]
onlyboxes ls = concatMap onlybox ls

cml :: (a -> [a]) -> [a] -> [a]
cml f ls = concatMap f ls

-- generic: main prover

-- all possible clauses
clauses :: (Logic f) => [f] -> [Clause f]
clauses [] = [Clause ([],[])]
clauses (a:as) = 
	let	padd c (Clause (p, n)) = Clause (c:p, n)
		nadd c (Clause (p, n)) = Clause (p, c:n)
	in	(map (padd a) (clauses as)) ++ (map (nadd a) (clauses as))
	
-- dnf
allsat :: (Logic f) => f -> [Clause f]
allsat phi = filter (\x -> (eval phi x)) (clauses (ma phi))

-- cnf
cnf :: (Logic f) => f -> [Clause f]
cnf phi = map (\(Clause (x, y)) -> (Clause (y, x))) (allsat (neg phi))

-- sat
sat :: (Logic f) => f -> Bool
sat phi = not $ provable $ neg phi

-- test
phi1 :: L1; phi1 =  (HM1 'a' (Atom1 0)) /\ (HM1 'a' (Atom1 1)) -->
  (HM1 'a' ((Atom1 0) /\ (Atom1 0)))

-- (pl, nl) = (c1, c2) where Clause (c1, c2) = cnf phi1
-- (pl, nl) = cnf phi1
pl = [HM1 'a' (And1 (Atom1 0) (Atom1 0))]
nl = [HM1 'a' (Atom1 0),HM1 'a' (Atom1 1)]
isag a (HM1 ind x) = (ind == a); -- is it agent a?
ag (HM1 ind x) = ind; -- find the agent
ags = nub (map ag (pl ++ nl)); -- the set of agents we have
aglits = map (\x -> (filter (isag x) pl, filter (isag x) nl)) ags;
-- this is a list of pairs (pl, nl) for each agent
-- strlits  = map (\(x, y) -> (map neg (striplst x), striplst y)) aglits;
-- the same with boxes stripped
-- comb (nls, pls) = map disjlst (map (\x -> x:nls) pls);
-- turn a combination pls/nls into a list of formulas
-- fllist = map comb strlits;
-- a list of lists consisting of formulas
--   in concat fllist
 
